/*----------------------------------------------------------------------------
 *      A R 1 6 6  -  K e r n e l
 *----------------------------------------------------------------------------
 *      Name:    AR166_CONFIG.C
 *      Purpose: Configuration of AR-166 Kernel by user.
 *      Rev.:    V1.02 / 11-oct-2004 
 *----------------------------------------------------------------------------
 *      This code is part of the AR-166 kernel package of Keil Software.
 *      Copyright (c) 2004 Keil Software. All rights reserved.
 *---------------------------------------------------------------------------*/

#pragma SAVE
#pragma OPTIMIZE(7) 
#pragma WARNING disable = 195           /* Disable C166 extended asm warning */

#include <AR166_Config.h>               /* AR166 user configuration header   */

/*----------------------------------------------------------------------------
 *      AR166 User configuration part BEGIN
 *---------------------------------------------------------------------------*/

//-------- <<< Use Configuration Wizard in Context Menu >>> -----------------
//
// <h>Task Definitions
// ===================
//
//   <o>OS_TASKCNT: Number of concurrent running tasks <0-250>
//   <i> Can be less than the number of defined tasks or more if
//   <i> multi instances of tasks will run. This define is used
//   <i> to calculate the the memory block size that is reserved
//   <i> for fixed memory block management.
//   <i> Default: 6
#define OS_TASKCNT      6

//   <o>OS_PRIVCNT: Number of tasks using private user stack <0-250>
//   <i> Some tasks may require bigger user stack than OS_STKSIZE.
//   <i> In this case the memory space for user stack must be
//   <i> provided by the user and is not assigned by the system.
//   <i> Default: 0
#define OS_PRIVCNT      0

//   <o>OS_STKSIZE: Task stack size in words <5-500>
//   <i> This stack is used for task's user stack and context switch
//   <i> for registers and system stack storage.
//   <i> Default: 100
#define OS_STKSIZE      100

//   <o>OS_TIMERCNT: Number of user timers <0-250>
//   <i> Define max. number of user timers running. They act as
//   <i> a watchdog timer counting down on each system clock tick.
//   <i> On timeout the user provided function is called.
//   <i> Default: 0  (User timers disabled)
#define OS_TIMERCNT     0

// </h>
// <h>System Timer Configuration
// =============================
//   <o>OS_TIMER: AR166 Kernel timer number <0=> Timer 0 <1=> Timer 1 <2=> Timer 2
//                                          <3=> Timer 3 <4=> Timer 4 <5=> Timer 5
//                                          <6=> Timer 6 <7=> Timer 7 <8=> Timer 8
//   <i> Define the C16x/XC16x timer used as a system tick timer
//   <i> Default: Timer 2
#define OS_TIMER        2

//   <o1>OS_CLOCK: Timer reload value <0-65535>
//   <i> Set the timer reload value for desired tick rate
//   <i> Default: 25000  (10ms at 20MHz and prescaler by 8)
#define OS_CLOCK        (0 - (unsigned int)2500)

// </h>
// <e>Round-Robin Task switching
#define OS_ROBIN        1

//   <o>OS_ROBINTOUT: Round-Robin Timeout in system ticks <1-1000>
//   <i> This is the time slice assigned to a running task.
//   <i> When this expires, round robin task switch takes place.
//   <i> Default: 5
#define OS_ROBINTOUT    5

// </e>

//   System Interrupt priority levels
//   ================================
//   ILVL: Interrrupt Level for system tick timer (0-15)
//       WARNING! The system timer interrupt level must be the LOWEST of all
//       other interrupts because otherwise the system timer would interrupt
//       other interrupt service routines and the context switch would fail.
//       Default: 1
#define ILVL            1

//   GLVL: Group Priority Level (0-3)
//       Set the Interrupt Group Level for selected System tick timer 
//       Default: 0
#define GLVL            0

//------------- <<< end of configuration section >>> -----------------------

/*----------------------------------------------------------------------------
 *      AR166 User configuration part END
 *---------------------------------------------------------------------------*/

#if   (OS_TIMER == 0)                   /* Timer 0                           */
  #define OS_TREG       T0
  #define OS_TOVF       T0IR
  #define OS_INTNR      0x20
  #define OS_TIE        T0IE
  #define OS_TRUN()     _bfld_ (T01CON,0xFF,0x40)
  #define OS_TIEN()     T0IC = aINT_ENABLE
  #define OS_TSET()     T0 = T0REL = OS_CLOCK
  #define OS_TREL()   
#elif (OS_TIMER == 1)                   /* Timer 1                           */
  #define OS_TREG       T1
  #define OS_TOVF       T1IR
  #define OS_INTNR      0x21
  #define OS_TIE        T1IE
  #define OS_TRUN()     _bfld_ (T01CON,0xFF00,0x4000)
  #define OS_TIEN()     T1IC = aINT_ENABLE
  #define OS_TSET()     T1 = T1REL = OS_CLOCK
  #define OS_TREL()   
#elif (OS_TIMER == 2)                   /* Timer 2                           */
  #define OS_TREG       T2
  #define OS_TOVF       T2IR
  #define OS_INTNR      0x22
  #define OS_TIE        T2IE
  #define OS_TRUN()     T2CON = 0x0040
  #define OS_TIEN()     T2IC = aINT_ENABLE
  #define OS_TSET()     T2 = OS_CLOCK
  #define OS_TREL()     T2 += OS_CLOCK 
#elif (OS_TIMER == 3)                   /* Timer 3                           */
  #define OS_TREG       T3
  #define OS_TOVF       T3IR
  #define OS_INTNR      0x23
  #define OS_TIE        T3IE
  #define OS_TRUN()     T3CON = 0x0040
  #define OS_TIEN()     T3IC = aINT_ENABLE
  #define OS_TSET()     T3 = OS_CLOCK
  #define OS_TREL()     T3 += OS_CLOCK
#elif (OS_TIMER == 4)                   /* Timer 4                           */
  #define OS_TREG       T4
  #define OS_TOVF       T4IR
  #define OS_INTNR      0x24
  #define OS_TIE        T4IE
  #define OS_TRUN()     T4CON = 0x0040
  #define OS_TIEN()     T4IC = aINT_ENABLE
  #define OS_TSET()     T4 = OS_CLOCK
  #define OS_TREL()     T4 += OS_CLOCK
#elif (OS_TIMER == 5)                   /* Timer 5                           */
  #define OS_TREG       T5
  #define OS_TOVF       T5IR
  #define OS_INTNR      0x25
  #define OS_TIE        T5IE
  #define OS_TRUN()     T5CON = 0x0041
  #define OS_TIEN()     T5IC = aINT_ENABLE
  #define OS_TSET()     T5 = OS_CLOCK
  #define OS_TREL()     T5 += OS_CLOCK
#elif (OS_TIMER == 6)                   /* Timer 6                           */
  #define OS_TREG       T6
  #define OS_TOVF       T6IR
  #define OS_INTNR      0x26
  #define OS_TIE        T6IE
  #define OS_TRUN()     T6CON = 0x8041
  #define OS_TIEN()     T6IC = aINT_ENABLE
  #define OS_TSET()     T6 = CAPREL = OS_CLOCK
  #define OS_TREL()   
#elif (OS_TIMER == 7)                   /* Timer 7                           */
  #define OS_TREG       T7
  #define OS_TOVF       T7IR
  #define OS_INTNR      0x3D
  #define OS_TIE        T7IE
  #define OS_TRUN()     _bfld_ (T78CON,0xFF,0x40)
  #define OS_TIEN()     T7IC = aINT_ENABLE
  #define OS_TSET()     T7 = T7REL = OS_CLOCK
  #define OS_TREL()
#elif (OS_TIMER == 8)                   /* Timer 8                           */
  #define OS_TREG       T8
  #define OS_TOVF       T8IR
  #define OS_INTNR      0x3E
  #define OS_TIE        T8IE
  #define OS_TRUN()     _bfld_ (T78CON,0xFF00,0x4000)
  #define OS_TIEN()     T8IC = aINT_ENABLE
  #define OS_TSET()     T8 = T8REL = OS_CLOCK
  #define OS_TREL()
#else
  #error OS_TIMER invalid
#endif


#define aINT_ENABLE     (0x40 | (ILVL << 2) | GLVL)
#define INITIAL_PSW     0x0800
#define Tsk_Lock()      OS_TIE = 0
#define Tsk_Unlock()    OS_TIE = 1

/*------------------------ Save Reduced Context Macro -----------------------*/

#define Save_Reduced_Ctx() __asm {                                     \
 __asm         MOV     [-R0],R13       /* Save Registers R13..R15   */ \
 __asm         MOV     [-R0],R14                                       \
 __asm         MOV     [-R0],R15                                       \
 __asm         MOV     R4,CP           /* Save Context Pointer (CP) */ \
 __asm         MOV     [-R0],R4                                        \
 __asm         MOV     R4,DPP3         /* Save Data Page Pointer 3  */ \
 __asm         MOV     [-R0],R4                                        \
 __asm         MOV     R4,DPP0         /* Save Data Page Pointer 0  */ \
 __asm         MOV     [-R0],R4                                        \
}

/*------------------------ Restore Reduced Context Macro --------------------*/

#define Restore_Reduced_Ctx() __asm {                                  \
 __asm         MOV     R4,[R0+]        /* Restore DPP0              */ \
 __asm         MOV     DPP0,R4                                         \
 __asm         MOV     R4,[R0+]        /* Restore DPP3              */ \
 __asm         MOV     DPP3,R4                                         \
 __asm         MOV     R4,[R0+]        /* Restore Context Pointer   */ \
 __asm         MOV     CP,R4                                           \
 __asm         NOP                                                     \
 __asm         MOV     R15,[R0+]       /* Restore registers R13..R15*/ \
 __asm         MOV     R14,[R0+]                                       \
 __asm         MOV     R13,[R0+]                                       \
}

/*------------------------ Save System Stack Macro --------------------------*/

#define Save_Sys_Stk(_Rg) __asm {                                      \
 __asm Scopy:                                                          \
 __asm         POP     R4              /* Copy word per word        */ \
 __asm         MOV     [-R0],R4                                        \
 __asm         CMPD1   _Rg,#1                                          \
 __asm         JMP     CC_NZ,Scopy     /* Copied in reverse order   */ \
}

/*------------------------ Restore System Stack Macro -----------------------*/

#define Restore_Sys_Stk(_Rg) __asm {                                   \
 __asm Tcopy:                                                          \
 __asm         MOV     R4,[R0+]        /* Copy word per word        */ \
 __asm         PUSH    R4                                              \
 __asm         CMPD1   _Rg,#1                                          \
 __asm         JMP     CC_NZ,Tcopy                                     \
}

/*------------------------ Return from interrupt ----------------------------*/

#define _reti_() __asm { RETI }

/*----------------------------------------------------------------------------
 *      Global Variables
 *---------------------------------------------------------------------------*/

extern P_TCB FAST_MEM os_runtask;
extern struct OS_XCB NEAR_MEM os_rdy;
extern struct OS_TCB NEAR_MEM os_clock_TCB;
extern WORD const sys_stktop;
extern WORD const sys_stkbot;
extern WORD const os_cnttask;
extern WORD const os_maxtaskn;
extern WORD NEAR_MEM os_time;
WORD const os_maxtaskrun = OS_TASKCNT;
/* Report uVision2 debugger user stack size in bytes */
WORD const os_stacksize  = OS_STKSIZE * 2;
WORD const os_timernum   = (OS_TIMERCNT << 8) | OS_TIMER;
WORD const os_clockrate  = OS_CLOCK; 
WORD const os_intlevel   = (GLVL << 8) | ILVL;

/*----------------------------------------------------------------------------
 *      Local Variables
 *---------------------------------------------------------------------------*/
/* Memory pool for TCB allocation    */
WORD MBOX_MEM m_tcb[(sizeof(struct OS_TCB) * OS_TASKCNT)/2 + 3];

/* Memory pool for User stack allocation. Need to allocate 2 additional  */
/* entries for 'os_clock_demon()' and 'os_idle_demon()'.                 */
WORD NEAR_MEM m_stk[OS_STKSIZE * (OS_TASKCNT-OS_PRIVCNT+2) + 3];

/* An array of Active task pointers.                                     */
P_TCB NEAR_MEM os_active_TCB[OS_TASKCNT];

/* Task switch has been requested from a task that has deleted itself.   */
BIT os_del_flag;

#if (OS_ROBIN == 1)
  WORD FAST_MEM  os_robin_time;
  P_TCB FAST_MEM os_tsk_robin;
#endif

#if (OS_TIMERCNT != 0)
  /* Memory pool for User Timer allocation  */
  WORD NEAR_MEM m_tmr[(sizeof(struct OS_TMR) * OS_TIMERCNT)/2 + 3];
#endif

/*----------------------------------------------------------------------------
 *      Global Functions
 *---------------------------------------------------------------------------*/

/*--------------------------- os_idle_demon ---------------------------------*/

void os_idle_demon (void) {
   /* The idle demon is a system task. It is running when no other task is   */
   /* ready to run (idle situation). It must not terminate. Therefore it     */
   /* should contain at least an endless loop.                               */
   for (;;) {
   /* HERE: include here optional user code to be executed when no task runs.*/
   _idle_ ();
   };
} /* end of os_idle_demon */


/*--------------------------- os_tmr_call -----------------------------------*/

void os_tmr_call (WORD info) {
   /* This function is called when the user timer has expired.               */
   /* Parameter "info" is the parameter defined when the timer was created.  */
   /* HERE: include here optional user code to be executed on timeout.       */
   info;
} /* end of os_tmr_call */

/*--------------------------- os_clock_interrupt ----------------------------*/

#pragma NOFRAME                         /* Do not store any C167 registers   */

void os_clock_interrupt (void)  interrupt OS_TIMER_INT = OS_INTNR  { 
   /* Do task switch to clock demon: entered upon a clock interrupt. */
   register WORD reg;

   __asm {                              ; Save Full Context
#ifndef __XC_DEVICE__
        PUSH    PSW                     ; chip bug, MULIP is not stored
#endif
        MOV     [-R0],R1                ; Save Registers R1..R15
        MOV     [-R0],R2
        MOV     [-R0],R3
        MOV     [-R0],R4
        MOV     [-R0],R5
        MOV     [-R0],R6
        MOV     [-R0],R7
        MOV     [-R0],R8
        MOV     [-R0],R9
        MOV     [-R0],R10
        MOV     [-R0],R11
        MOV     [-R0],R12
        MOV     [-R0],R13
        MOV     [-R0],R14
        MOV     [-R0],R15
        MOV     R4,CP                   ; Save Context Pointer (CP)
        MOV     [-R0],R4
        MOV     R4,DPP3                 ; Save Data Page Pointer 3
        MOV     [-R0],R4
        MOV     R4,DPP0                 ; Save Data Page Pointer 0
        MOV     [-R0],R4
        MOV     R4,#0
        JNB     MDRIU,Lab1              ; Save Mul/Div Registers
        SCXT    MDC,#10H                ; only if required
        MOV     R4,MDH                  ; Save Mul/Div High
        MOV     [-R0],R4
        MOV     R4,MDL                  ; Save Mul/Div Low
        MOV     [-R0],R4
        POP     R4                      ; Save Mul/Div Control
   Lab1:
        MOV     [-R0],R4                ; Store 1 or 3 words
   }

#ifdef __XC_DEVICE__
      __asm {                           ; Store MAC registers of XC167
        MOV     R4,MSW                  ; Store MAC status word
        MOV     [-R0],R4
        MOV     R4,MAL                  ; Store MAC acumulator Low word
        MOV     [-R0],R4
        MOV     R4,MAH                  ; Store MAC acumulator High word
        MOV     [-R0],R4
        SCXT    MRW,#00H                ; Store & Clear MAC repeat word
        POP     R4
        MOV     [-R0],R4
        MOV     R4,IDX0                 ; Store MAC Address pointer 0
        MOV     [-R0],R4
   }
#endif

   reg = (sys_stktop - SP) >> 1;
   os_runtask->sys_load = reg;
   if (reg) {
      Save_Sys_Stk(reg);
   }
   __asm {
        MOV     reg,R0
   }
   os_runtask->usr_stack = reg;
   reg = os_clock_TCB.usr_stack;
   __asm {
        MOV     R0,reg
   }
   reg = os_clock_TCB.sys_load;
   if (reg) {
      Restore_Sys_Stk(reg);
   }
   Restore_Reduced_Ctx();
   Tsk_Lock ();
   os_runtask->full_ctx = TRUE;
   os_runtask->state  = READY;
   os_put_rdy_first (os_runtask);
   os_runtask = &os_clock_TCB;
   os_clock_TCB.state = RUNNING;
} /* end of os_clock_interrupt */

/*--------------------------- os_tmr_init -----------------------------------*/

void os_tmr_init (void) {
   /* Initialize hardware timer as system tick timer. This function is       */
   /* called at the system startup.                                          */
   OS_TSET();
   OS_TIEN(); 
   OS_TRUN();
#if (OS_ROBIN == 1)
   os_robin_time = OS_ROBINTOUT;
#endif
} /* end of os_tmr_init */

/*--------------------------- os_tmr_reload ---------------------------------*/

void os_tmr_reload (void) {
   /* Reload system timer for next period if required.                       */
   OS_TREL();
} /* end of os_tmr_reload */

/*--------------------------- os_tmr_set_ovf --------------------------------*/

void os_tmr_set_ovf (void) {
   /* Set timer overflov flag.                                               */
   OS_TOVF = 1;
} /* end of os_tmr_set_ovf */

/*--------------------------- os_tmr_inspect_cnt ----------------------------*/

WORD os_tmr_inspect_cnt (void) {
   /* Inspect current value of rtx timer.                                    */
   return (OS_TREG);
} /* end of os_tmr_inspect_cnt */

/*--------------------------- os_tmr_inspect_ovf ----------------------------*/

BIT os_tmr_inspect_ovf (void) {
   /* Inspect current state of timer overflow flag.                          */
   return (OS_TOVF);
} /* end of os_tmr_inspect_ovf */

/*--------------------------- tsk_lock --------------------------------------*/

void tsk_lock (void) {
   /* Lock out tasks: prevents task switching by locking out scheduler       */
   /* activation on interrupt. 'ATOMIC' is needed to avoid pipeling effect.  */

   __asm {
#if (OS_TIMER < 7)
        ATOMIC  #3                      ; to avoid pipeling effect
#else
        EXTR    #3                      ; to avoid pipeling effect
#endif
        BCLR    OS_TIE
        NOP
        NOP
   }
} /* end of tsk_lock */

/*--------------------------- tsk_unlock ------------------------------------*/

void tsk_unlock (void) {
   /* Enable AR System Tick Timer Interrupts.                                */
   OS_TIE = 1;
} /* end of tsk_unlock */

/*--------------------------- int_disable -----------------------------------*/

void int_disable (void) {
   /* This function is used to disable all interrupts during AR166 kernel    */
   /* startup when the system is beeing initialized and configured.          */
   __asm {
        ATOMIC  #3                      ; to avoid pipeling effect
        BCLR    IEN
        NOP
        NOP
   }
} /* end of int_disable */

/*--------------------------- int_enable ------------------------------------*/

void int_enable (void) {
   /* This function is used to enable global interrupts when the AR166       */
   /* has properly initialized and started.                                  */
   IEN = 1;
} /* end of int_enable */

/*--------------------------- os_init_mem -----------------------------------*/

void os_init_mem (void) {
   WORD i;

   for (i = 0; i < OS_TASKCNT; i++) {
      os_active_TCB[i] = NULL;
   }
   os_del_flag = FALSE;
   _init_box  (&m_tcb, sizeof(m_tcb), sizeof(struct OS_TCB));
   _init_boxn (&m_stk, sizeof(m_stk), OS_STKSIZE*2);
#if (OS_TIMERCNT != 0)
   _init_boxn (&m_tmr, sizeof(m_tmr), sizeof(struct OS_TMR));
#endif
} /* end of os_init_mem */

/*--------------------------- os_alloc_TCB ----------------------------------*/

P_TCB os_alloc_TCB () {
   return (_alloc_box (m_tcb));
} /* end of os_alloc_TCB */

/*--------------------------- os_free_TCB -----------------------------------*/

void os_free_TCB (P_TCB p_TCB) {
   /* Free allocated memory resources for the task "p_TCB" */
   if (p_TCB == os_runtask) {
      /* The running task has deleted itself. */
      os_del_flag = TRUE;
   }
   _free_boxn (m_stk, p_TCB->stack);
   _free_box  (m_tcb, (void MBOX_MEM *)p_TCB);
} /* end of os_free_TCB */

/*--------------------------- os_alloc_TMR ----------------------------------*/

P_TMR os_alloc_TMR () {
#if (OS_TIMERCNT != 0)
   return (_alloc_boxn (m_tmr));
#else
   return (PNULL);
#endif
} /* end of os_alloc_TMR */

/*--------------------------- os_free_TMR -----------------------------------*/

void os_free_TMR (P_TMR timer) {
   /* Free allocated memory resources for user timer 'timer' */
#if (OS_TIMERCNT != 0)
   _free_boxn (m_tmr, timer);
#else
   timer;
#endif
} /* end of os_free_TMR */

/*--------------------------- os_init_context -------------------------------*/

void os_init_context (P_TCB p_TCB,
                      BYTE  priority,
                      FUNCP task_body,
                      BYTE  full_contxt) {
   WORD NEAR_MEM *stk,i;

   /* Prepare TCB and saved context for a first time start of a task         */
   /* "p_TCB" points to TCB to be initialised. "priority" indicates desired  */
   /* execution priority. "task_body" is the start address of the task.      */
   /* "full_context" identifies context type.                                */
   p_TCB->cb_type = TCB;              /* Initialise general part of TCB */
   p_TCB->state   = READY;
   p_TCB->prio    = priority;
   p_TCB->p_lnk   = NULL;
   p_TCB->p_rlnk  = NULL;
   p_TCB->p_dlnk  = NULL;
   p_TCB->p_blnk  = NULL;
   p_TCB->delta_time    = 0;
   p_TCB->interval_time = 0;
   p_TCB->events  = 0;
   p_TCB->waits   = 0;

   /* Initialise C167 specific part of TCB */
   p_TCB->full_ctx = full_contxt;
  
   /* Prepare a complete interrupt frame for first task start            */
   if (p_TCB->priv_stack != 0) {
      /* User has provided a memory space for the User stack. */
      stk = &p_TCB->stack[p_TCB->priv_stack>>1];
   }
   else {
      /* Allocate the memory space for the User stack. */
      p_TCB->stack = _alloc_boxn (m_stk);
      stk = &p_TCB->stack[OS_STKSIZE]; /* Write to the top of stack.     */
   }
   for (i = full_contxt ? 15 : 3; i; i--) {  /* Write initial registers  */
      *--stk = 0;
   } 
   *--stk = CP;                       /* Only default supported for now. */
   *--stk = DPP3;
   *--stk = DPP0;
   i = 0;
   if (full_contxt) {   
      *--stk = 0;                     /* Mul/Div default not stored      */
#ifdef __XC_DEVICE__
      *--stk = 0;                     /* Initial MSW                     */
      *--stk = 0;                     /* Initial MAL                     */
      *--stk = 0;                     /* Initial MAH                     */
      *--stk = 0;                     /* Initial MRW                     */
      *--stk = 0;                     /* Initial IDX0                    */
#else
      *--stk = INITIAL_PSW;           /* Initial PSW - C167 chip bug     */
      i++;
#endif

   }
   *--stk = (WORD)task_body;          /* Initial IP                      */

#ifndef __SEGMENTED_MODEL__
   *--stk = INITIAL_PSW;              /* Initial PSW for RETI            */
   i += 2;
#else
   *--stk = (WORD)((DWORD)task_body>>16); /* Initial CSP Segmented mode  */
   *--stk = INITIAL_PSW;              /* Initial PSW for RETI            */
   i += 3;
#endif

   p_TCB->sys_load  = i;
   p_TCB->usr_stack = (WORD)stk;      /* Initial User stack              */
} /* end of os_init_context */


/*--------------------------- os_set_env ------------------------------------*/

void os_set_env (P_TCB p_TCB) {
   /* Fix up runtime environment to fit idle task. It is called after the  */
   /* idle task TCB initialization. "p_TCB" identifies the TCB to be used. */
   register WORD NEAR_MEM *usp;

   usp = &p_TCB->stack[OS_STKSIZE];
   __asm {
        POP     R1                      ; IP
#ifndef __FUNC_MODEL_NEAR__
        POP     R2                      ; CSP
#endif
   }
   __asm {
        ATOMIC  #2                      ; disable inter. for next 2 instr.
        MOV     SP,WORD sys_stktop      ; reinit system stack
        ;                               ; next instruction must not use SP
        MOV     R0,usp                  ; setup the user stack pointer
#ifndef __FUNC_MODEL_NEAR__
        PUSH    R2                      ; CSP
#endif
        PUSH    R1                      ; IP
   }
} /* end of os_set_env */


/*--------------------------- os_switch_tasks -------------------------------*/

void FAR_MEM os_switch_tasks (P_TCB p_new) {
   /* Switch to next task (identified by "p_new"): save old and restore new  */
   /* context, use far call to obtain full address. 'p_new' assigned to R7.  */
   /* Do not save Task context if this is a call from a task that has        */
   /* deleted itself. The 'os_task_TCB' has already been deallocated.        */
   register WORD reg1,reg2;
 
   if (!os_del_flag) {
      os_runtask->full_ctx = FALSE;
      Save_Reduced_Ctx();

#ifndef __SEGMENTED_MODEL__
      __asm {                           ; Correct Stack content for RETI
        POP     R1                      ; Pop IP
        PUSH    PSW
        PUSH    R1                      ; Push IP
      }
#else
      __asm {                           ; Correct Stack content for RETI
        POP     R1                      ; Pop IP
        POP     R2                      ; Pop CSP
        PUSH    PSW
        PUSH    R2                      ; Push CSP
        PUSH    R1                      ; Push IP
      }
#endif
      reg1 = (sys_stktop - SP) >> 1;
      os_runtask->sys_load = reg1;
      if (reg1) {
         Save_Sys_Stk(reg1);
      }
      __asm {
        MOV     reg1,R0
      }
      os_runtask->usr_stack = reg1;
   }
   else {
      /* Reinit system stack. */
      __asm {
        ATOMIC  #2
        MOV     SP,WORD sys_stktop      ; reinit system stack
        ;                               ; next instruction must not use SP
      }
      os_del_flag = FALSE;
   }
   reg2 = p_new->usr_stack;
   __asm {
        MOV     R0,reg2
   }
   reg2 = p_new->sys_load;
   if (reg2) {
      Restore_Sys_Stk(reg2);
   }
   os_runtask = p_new;
   p_new->state = RUNNING;
#if (OS_ROBIN == 1)
   os_tsk_robin = p_new;
#endif
   if (p_new->full_ctx) {

#ifdef __XC_DEVICE__
      __asm {                           ; Restore MAC registers of XC167
        MOV     R4,[R0+]                ; Restore MAC Address pointer 0
        MOV     IDX0,R4
        MOV     R4,[R0+]                ; Restore MAC repeat word
        MOV     MRW,R4
        MOV     R4,[R0+]                ; Restore MAC acumulator High word
        MOV     MAH,R4
        MOV     R4,[R0+]                ; Restore MAC acumulator Low word
        MOV     MAL,R4
        MOV     R4,[R0+]                ; Restore MAC status word
        MOV     MSW,R4
      }
#endif

      __asm {                           ; Restore Full Context
        MOV     R4,[R0+]                ; Get MDC register or flag
        JMPR    CC_Z,Lab1               ; Flag was 0, no restoring
        MOV     R3,[R0+]                ; Restore Mul/Div Regs
        MOV     MDL,R3
        MOV     R3,[R0+]
        MOV     MDH,R3
        MOV     MDC,R4
   Lab1:
        MOV     R4,[R0+]                ; Restore Data Page Pointer 0
        MOV     DPP0,R4
        MOV     R4,[R0+]                ; Restore Data Page Pointer 3
        MOV     DPP3,R4
        MOV     R4,[R0+]                ; Restore Context Pointer
        MOV     CP,R4
        NOP                             ; This instr. must not access registers
        MOV     R15,[R0+]               ; Restore registers R1..R15
        MOV     R14,[R0+]
        MOV     R13,[R0+]
        MOV     R12,[R0+]
        MOV     R11,[R0+]
        MOV     R10,[R0+]
        MOV     R9,[R0+]
        MOV     R8,[R0+]
        MOV     R7,[R0+]
        MOV     R6,[R0+]
        MOV     R5,[R0+]
        MOV     R4,[R0+]
        MOV     R3,[R0+]
        MOV     R2,[R0+]
        MOV     R1,[R0+]
#ifndef __XC_DEVICE__
        POP     PSW
#endif
      }
      Tsk_Unlock ();
      _reti_();
      }
   else {
      Restore_Reduced_Ctx();
      Tsk_Unlock ();
      _reti_();
      };
} /* end of os_switch_tasks */


/*--------------------------- os_chk_robin ----------------------------------*/

void os_chk_robin (void) {
   /* Check if Round Robin timeout expired and switch to the next ready task.*/
   /* This function is called from the "os_clock_demon()" task scheduler.    */
#if (OS_ROBIN == 1)
   P_TCB p_new;

   if (os_rdy.p_lnk != os_tsk_robin) {
      os_robin_time == os_time + OS_ROBINTOUT;
      return;
      }
   if (os_robin_time == os_time) {
      /* Round Robin timeout has expired. */
      os_robin_time += OS_ROBINTOUT;
      p_new = os_get_first (&os_rdy);
      os_put_prio ((P_XCB)&os_rdy, p_new);
      }
#endif
} /* end of os_chk_robin */


#pragma RESTORE

/*----------------------------------------------------------------------------
 * end of file
 *---------------------------------------------------------------------------*/

