/*----------------------------------------------------------------------------
 *      A R 1 6 6   R o u n d   R o b i n   E x a m p l e
 *----------------------------------------------------------------------------
 *      Name:    RoundRobin.C
 *      Purpose: Demonstration of RoundRobin Task switching
 *---------------------------------------------------------------------------*/

#include <ar166.h>                    /* AR166 kernel functions & defines    */

/* id1, id2 will contain task identifications at run-time */
OS_TID id1, id2, id3;
unsigned long counter1;
unsigned long counter2;
unsigned long counter3;
 
/*----------------------------------------------------------------------------
 *   Task 1 'task1':  ARTX-166 Kernel starts this task with os_sys_init (1)
 *---------------------------------------------------------------------------*/
void task1 (void) _task_ 1 _priority_ 2 {
  /* Obtain own system task identification number */
  id1 = os_tsk_self ();
  /* Assign system identification number of task2 to id2 */
  id2 = os_tsk_create (2);
  id3 = os_tsk_create (3);
  while (1)  {                        /* endless loop                         */
    counter1++;                       /* increment counter 1                  */
  }
}
 
/*----------------------------------------------------------------------------
 *   Task 2 'task2':  ARTX-166 Kernel starts this task with os_tsk_create (2)
 *---------------------------------------------------------------------------*/
void task2 (void) _task_ 2 _priority_ 2 {
  while (1)  {                        /* endless loop                         */
    counter2++;                       /* increment counter 2                  */
    if ((counter2 & 0xFFFF) == 0) os_evt_set (0x0004, id3);
  }
}

/*----------------------------------------------------------------------------
 *   Task 3 'task3':  ARTX-166 Kernel starts this task with os_tsk_create (3)
 *---------------------------------------------------------------------------*/
void task3 (void) _task_ 3 _priority_ 4 {
  while (1)  {                        /* endless loop                         */
    os_evt_wait_or (0x0004, 0xffff);
    counter3++;                       /* increment counter 1                  */
  }
}
 
/*----------------------------------------------------------------------------
 *        Main: Initialize and start ARTX Kernel
 *---------------------------------------------------------------------------*/
void main (void) {
  os_sys_init (1);
}

/*----------------------------------------------------------------------------
 * end of file
 *---------------------------------------------------------------------------*/


